#!/usr/bin/python
import sh
import sys
from collections import defaultdict
import bz2


# This is generator that converts stdin into chunks of size buffer_size,
# because we do not want to keep everything in memory
def data_generator(f):
    global buffer_size
    while True:
        size = buffer_size
        chunks = []
        while size:
            x = f.read(size)
            size -= len(x)
            if not x:
                yield b"".join(chunks) 
                return
            chunks.append(x)
        yield b"".join(chunks)

def shorten(data):
    global patterns, buffer_size
    haystack = data.next()
    have_data = True
    output = b""

    while True:
        if len(haystack) < buffer_size:
            # Add more data
            try:
                x = data.next()
            except StopIteration:
                x = b""
                have_data = False
            haystack += x
        # Seach for the nearest block
        block_positions = filter(
            lambda x: x != -1,
            [haystack.find(pattern) for pattern in patterns]
        )
        if len(block_positions) == 0:
            if have_data:
                # There is no block in haystack, store it to the output, keep
                # few bytes (compressed magic could be split by buffering) 
                output += haystack[:-6]
                haystack = haystack[-6:]
                continue
            else:
                # No block, no data, just quit
                break
        # Output everything up to the block
        nearest_block = min(block_positions)
        output += haystack[:nearest_block]
        haystack = haystack[nearest_block:]
        # Remove all possible repetitions
        needle = haystack[:200] 
        period = haystack.find(needle, 1)
        if period != -1 and haystack[:period] == haystack[period:2 * period]:
            repeat = haystack[:period]
            del_count = 0
            count = 1
            while haystack[period * count:period * (count + 1)] == repeat:
                if (count+3) * period > len(haystack):
                    haystack = haystack[period * count:]
                    count = 0
                count += 1
                del_count += 1
                if len(haystack) < buffer_size:
                    try:
                        x = data.next()
                    except StopIteration:
                        have_data = False
                        x = b""
                    haystack += x
            haystack = haystack[period * count:]
        else:
            # No period (of length at most buffer_size), skip this block
            output += haystack[:6]
            haystack = haystack[6:]
    output += haystack
    return output

def unpack(data):
    global buffer_size
    if type(data) != str:
        data = b"".join(data)
    for x in sh.bunzip2(
        _in=data,
        _ok_code=[0, 2], # ignore errors
        _out_bufsize=buffer_size, # we want buffered output
    ):
        yield x
    
if __name__ == '__main__':
    if len(sys.argv) < 2:
        print "Usage: python solution-mic.py {easy.in,hard.in}"
        exit(1)

    # Compute all bit-shifts of compressed magic
    patterns = [
        '\x31\x41\x59\x26\x53\x59',
    ]
    for n in range(1, 8):
        A = [
            ((ord(x1) << n) & 0xFF) | (ord(x2) >> (8-n)) 
            for x1, x2 in zip(patterns[0], patterns[0][1:])
        ]
        patterns.append(b''.join(map(chr, A)))

    # Uncompress file twice if necessary
    buffer_size = 10000000
    input_file = sys.argv[1]
    if input_file == 'hard.in':
        twice_compressed = unpack(unpack(open(input_file, 'r')))
    else:
        twice_compressed = data_generator(open(input_file, 'r'))
    
    # Shorten, unpack, shorten, unpack
    unfiltered = b"".join(unpack(shorten(unpack(shorten(twice_compressed)))))
    filtered = unfiltered.replace(b"\x00", "")
    print filtered
